/*   FILE: dmapd-transcode.c -- Perform non-real-time transcoding. Even
 *         though dmapd can do this itself, a separate tool might be desireable if
 *         dmapd itself is running on a slow computer.
 * AUTHOR: W. Michael Petullo <mike@flyn.org>
 *   DATE: 13 June 2013
 *
 * Copyright (c) 2013 W. Michael Petullo <new@flyn.org>
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <config.h>
#include <glib.h>
#include <stdlib.h>
#include <string.h>

#include "dmapd-dmap-container-record.h"
#include "dmapd-dmap-container-db.h"
#include "dmapd-daap-record-factory.h"
#include "db-builder.h"
#include "util.h"
#include "util-gst.h"

#define DEFAULT_DB_MOD             "ghashtable"
#define DEFAULT_AV_META_READER_MOD "gst"

static GMainLoop *_loop;

static GSList  *_music_dirs            = NULL;
static GSList  *_music_formats         = NULL;
static gchar   *_module_dir            = NULL;
static gchar   *_output_dir            = NULL;
static gchar   *_transcode_mimetype    = NULL;
static gchar   *_db_module             = NULL;
static gchar   *_av_meta_reader_module = NULL;
static gboolean _enable_version        = FALSE;
static gboolean _enable_debug          = FALSE;

static void
_free_globals (void)
{
	g_debug ("Free'ing globals");

	g_slist_free_full(_music_dirs, g_free);
	g_slist_free_full(_music_formats, g_free);

	g_free (_output_dir);
	g_free (_transcode_mimetype);
}

static gboolean
_add_to_opt_list (const gchar *option_name,
                  const gchar *value,
                  G_GNUC_UNUSED gpointer data,
                  G_GNUC_UNUSED GError **error)
{
	if (strcmp (option_name, "-m") == 0) {
		_music_dirs = g_slist_append (_music_dirs, g_strdup (value));
	} else if (strcmp (option_name, "-M") == 0) {
		_music_formats = g_slist_append (_music_formats, g_strdup (value));
	} else {
		g_error ("Unsupported option: %s.", option_name);
	}
	return TRUE;
}

/* FIXME: how to enumerate available transcoding formats? */
static GOptionEntry _entries[] = {
	{ "music-dir", 'm', 0, G_OPTION_ARG_CALLBACK, _add_to_opt_list, "Music directory", NULL },
	{ "music-format", 'M', 0, G_OPTION_ARG_CALLBACK, _add_to_opt_list, "Acceptable music format", NULL },
	{ "output-dir", 'o', 0, G_OPTION_ARG_STRING, &_output_dir, "Output directory", NULL },
	{ "transcode-mimetype", 't', 0, G_OPTION_ARG_STRING, &_transcode_mimetype, "Target MIME type for transcoding", NULL },
	{ "version", 'v', 0, G_OPTION_ARG_NONE, &_enable_version, "Print version number and exit", NULL },
	{ NULL }
};

static gboolean
_transcode (DmapRecordFactory *factory,
            GSList *media_dirs,
            GSList *acceptable_formats,
            GError **error)
{
	gboolean ok = FALSE;
	GSList *l;
	DmapDb *db                    = NULL;
	DmapContainerDb *container_db = NULL;
	DbBuilder *builder            = NULL;

	g_assert (_db_module);

	db = DMAP_DB (util_object_from_module (TYPE_DMAPD_DMAP_DB, 
	                                  _module_dir,
					  _db_module,
					  "record-factory",
					  factory,
					  NULL));
	g_assert (db);

	if (acceptable_formats) {
		g_object_set (db, "acceptable-formats", acceptable_formats, NULL);
	}

	container_db = DMAP_CONTAINER_DB (dmapd_dmap_container_db_new ());
	builder = DB_BUILDER (util_object_from_module (TYPE_DB_BUILDER, _module_dir, "gdir", NULL));

	g_object_set (builder, "db", db, NULL);

	for (l = media_dirs; l; l = l->next) {
		ok = db_builder_build_db_starting_at (builder, l->data, NULL, error);
		if (!ok) {
			goto done;
		}
	}

	dmap_db_foreach (db, (DmapIdRecordFunc) util_gst_transcode_cache, &(db_dir_and_target_transcode_mimetype_t) { _output_dir, _transcode_mimetype });

done:
	if (NULL != db) {
		g_object_unref (db);
	}

	if (NULL != container_db) {
		g_object_unref (container_db);
	}

	if (NULL != builder) {
		g_object_unref (builder);
	}

	return ok;
}

static void
_sigterm_handler (G_GNUC_UNUSED int i)
{
	signal (SIGTERM, _sigterm_handler);

	g_debug ("Received TERM signal");

	g_main_loop_quit (_loop);
}

static const gchar *
_level_to_str(GLogLevelFlags level)
{
	const gchar *str;

	switch (level) {
	case G_LOG_LEVEL_ERROR:
		str = "ERROR";
		break;
	case G_LOG_LEVEL_CRITICAL:
		str = "CRITICAL";
		break;
	case G_LOG_LEVEL_WARNING:
		str = "WARNING";
		break;
	case G_LOG_LEVEL_MESSAGE:
		str = "MESSAGE";
		break;
	case G_LOG_LEVEL_INFO:
		str = "INFO";
		break;
	case G_LOG_LEVEL_DEBUG:
		str = "DEBUG";
		break;
	default:
		str = "UNKNOWN";
		break;
	}

	return str;
}

static void
_log_printf(const char *log_domain,
            GLogLevelFlags level,
            const gchar *message,
            G_GNUC_UNUSED gpointer user_data)
{
	if (level > G_LOG_LEVEL_INFO && !_enable_debug) {
		return;
	}

	g_printerr("%s: %s: %s\n",
	            log_domain,
	            _level_to_str(level & G_LOG_LEVEL_MASK),
	            message);
}

int
main (int argc, char *argv[])
{
	gboolean ok;
	int exitval = EXIT_SUCCESS;
	GError *error = NULL;
	GOptionContext *context;
	AvMetaReader *av_meta_reader = NULL;

	if (getenv ("DMAPD_DEBUG") != NULL) {
		_enable_debug = TRUE;
	}

	util_stringleton_init ();

	_loop = g_main_loop_new (NULL, FALSE);

	g_log_set_handler ("libdmapsharing", G_LOG_LEVEL_DEBUG, _log_printf, NULL);
	g_log_set_handler (NULL, G_LOG_LEVEL_DEBUG, _log_printf, NULL);

	// These must appear before parsing command-line options because
	// they contribute to which options are available.
	_module_dir = getenv ("DMAPD_MODULEDIR");
	_module_dir = _module_dir ? _module_dir : DEFAULT_MODULEDIR;
	
	_av_meta_reader_module = getenv ("DMAPD_AV_META_READER_MODULE");
	_av_meta_reader_module = _av_meta_reader_module ? _av_meta_reader_module : DEFAULT_AV_META_READER_MOD;

	_db_module = getenv ("DMAPD_DB_MODULE");
	_db_module = _db_module ? _db_module : DEFAULT_DB_MOD;

	context = g_option_context_new ("-m: transcode media");
	g_option_context_add_main_entries (context, _entries, NULL);

	if (strcmp (_av_meta_reader_module, "null") != 0) {
		GOptionGroup *group;
		av_meta_reader = AV_META_READER (util_object_from_module (TYPE_AV_META_READER,
		                                                     _module_dir,
								     _av_meta_reader_module,
								     NULL));
		if (av_meta_reader) {
			group = av_meta_reader_get_option_group (av_meta_reader);
			if (group) {
				g_option_context_add_group (context, group);
			}
		}
	}

	if (! g_option_context_parse (context, &argc, &argv, &error)) {
		g_error ("Option parsing failed: %s", error->message);
	}

	if (_enable_version) {
		g_print ("dmapd version %s\n", VERSION);
		exit (EXIT_SUCCESS);
	}

	if (! av_meta_reader) {
		g_error ("Need an AV metadata reader plugin");
	}

	if (! _music_dirs) {
		g_error ("Need to provide a music directory");
	}

	if (! _output_dir) {
		_output_dir = g_strdup (".");
	}

	g_option_context_free (context);

	signal (SIGTERM, _sigterm_handler);

	if (av_meta_reader == NULL) {
		g_error ("Music directory specified but AV metadata reader module is 'null'");
	}

	DmapRecordFactory *factory;
	factory = DMAP_RECORD_FACTORY (
			g_object_new (
				TYPE_DMAPD_DMAP_RECORD_FACTORY,
				"meta-reader",
				av_meta_reader,
				NULL));

	ok = _transcode(factory, _music_dirs, _music_formats, &error);
	if (!ok) {
		g_error("Error transcoding: %s", error->message);
	}

	if (av_meta_reader) {
		g_object_unref (av_meta_reader);
	}

	_free_globals ();
	util_stringleton_deinit ();
	g_debug ("Parent Exiting");

	exit(exitval);
}
